# Kaskade-S/T 
# v 1.0: 2010-12-30
# Encryption and decryption with a cascade of one substitution and one transposition
#  - Here it doesn't matter which encryption method is applied first.
#  - All input from files (despite the options).
#  - Key file can be generated using two random or two identical permutations.
#  - Length of a permutation is also the length of the alphabet. Default is 256.
#    If you specify a shorter length and a given offset (using -f) for the alphabet,
#    then the input files (plaintext or ciphertext) may ONLY contain characters
#    of the specified alphabet. E.g. if you use only the 26 uppercase letters:
#      python Kaskade.py  -g Random-Different  -l 26
#      python Kaskade.py  -e plaintext ciphertext  Key  -l 26  -f A
#      python Kaskade.py  -d ciphertext  plaintext-dec  Key  -l 26  -f A


### Used Functions
### ==============

# Invert a permutation (the key is a combination of two permutations of length lop)
#   Input:  length k of one permutation, given permutation
#   Output: inverted permutation
def InversePerm(k, perm_in):
  perm_out=perm_in[:]
  for i in range(k):   
     perm_out[perm_in[i]]=i 
  return perm_out


# Create a key
#   Input:  length k, bool RandomPermKey (True = random perm; False = identical perm)
#   Output: key
def CreatePermKey(k, RandomPermKey):
  if RandomPermKey == False:
    liste=[]
    for i in range(k):
      liste.append(i%k)
    return liste 
  else:
    return random.sample(range(k),k)


# Perform substitution encryption or decryption
# - cipher applied to one data block of length k,
#   Input:  block length k, perm key, data block
#   Output: substituted data block
def EncDecSubst(k, perm, DataIn, DataOut):
  for i in range(k):
    DataOut[i]=perm[DataIn[i]]


# Perform transposition encryption or decryption
# - cipher applied to one data block of length k,
#   Input:  block length k, perm key, data block 
#   Output: transpositioned data block
def EncDecTranspos(k, perm, DataIn, DataOut):
  for i in range(k):
    DataOut[perm[i]]=DataIn[i] 



### Default parameter settings
### ==========================

T_verbose = False             # If True: Show some intermediate results too (option -v).
T_firstAlphChar = False	      # True if you specify the smallest character of the alphabet.
			      # Use case e.g.: Alphabet is capital letters only. (True: when -f).
first=0		              # Specify the first letter of a shortend alphabet (must be 0 as default).
lop=256                       # length of one permutation (can be made smaller via option -l)
                              # The whole key exists of two permutations.
                              # Later on, data blocks are handled in the size of lop.
T_ReadPermKeysFromFile=False  # True: when (-e or -d) and 3 parameters are given
T_WritePermKeysToFile=False   # True: when -g
T_RandomPermKey=False         # True: when -g and key random (instead of identical perm)
T_SamePermKeys=False          # True: when -g and key parts are different (instead of two same perms)
T_EncryptFileArg=False        # True: Encryption (-e)
T_DecryptFileArg=False        # True: Decryption (-d)

Name_f_permout = "Key"  # The two permutations created are written into key file (used only when -g).



### Imports
### =======

import random
import os



### Handle Options (command line parameters)
### ========================================

## Get options

from optparse import OptionParser
pars = OptionParser(version='%prog version 1.0, Dec. 2010 (open source, written for the MTC3 contest)')

pars.add_option("-e", "--encrypt", help='Encrypts file-in, builds file_out, uses keyfile_in as key.',\
                dest='T_EncryptFileArg', default=False, action='store',\
                metavar='<file_in> <file_out> <keyfile_in>', nargs=3, type='string')

pars.add_option("-d", "--decrypt", help='Decrypts file-in, builds file_out, uses keyfile_in as key.',\
                dest='T_DecryptFileArg', default=False, action='store',\
                metavar='<file_in> <file_out> <keyfile_in>',  nargs=3, type='string')

pars.add_option("-g", "--genkey", help='Generates a key and stores it in key_out. There are 4 possible arguments to specify the generated key: Random-Different, Random-Same, Ident-Different, Ident-Same',\
                dest='T_genkey', default=False, action='store',\
                metavar='<RandomOrNot-DiffOrSame>',  nargs=1, type='string')

pars.add_option("-l", "--keylen", help='Defines key length. Default is 256. Can be made shorter.', type='int') 

pars.add_option("-v", "--verbose", help='Print commandline options, cwd, etc.',\
                dest='T_verbose', default=False, action='store_true')

pars.add_option("-f", "--first", help='Defines the first character of an alphabet shorter than 256 starting at the given character.', type='str') 

(opts, args) = pars.parse_args()


## Interpret options, set according global variables

# Verbose: -v
if opts.T_verbose:
  T_verbose = True
  print("CWD =", os.getcwd())
  print( "Opts:", opts )

# Specify the first letter of a shortend alphabet: -f
if opts.first:
  if opts.keylen:
    first = ord(opts.first[0])
    T_firstAlphChar = True
  else:
    print("Error: Option -l is missing.")
    exit()

# Encryption and decryption: -e and -d
if opts.T_EncryptFileArg or opts.T_DecryptFileArg:
  T_ReadPermKeysFromFile=True
  
  if opts.T_EncryptFileArg:
    T_EncryptFileArg=True
    Name_f_in  = opts.T_EncryptFileArg[0]
    Name_f_out = opts.T_EncryptFileArg[1]
    Name_f_permin = opts.T_EncryptFileArg[2]    

  if opts.T_DecryptFileArg:
    T_DecryptFileArg=True
    Name_f_in  = opts.T_DecryptFileArg[0]
    Name_f_out = opts.T_DecryptFileArg[1]
    Name_f_permin = opts.T_DecryptFileArg[2]

  if not os.path.exists(Name_f_in):
    print("Error: File \"", Name_f_in, "\" does not exist.")
    exit()

  if not os.path.exists(Name_f_permin):
    print("Error: File \"", Name_f_permin, "\" does not exist.")
    exit()

# Generate key: -g
if opts.T_genkey:
  T_WritePermKeysToFile=True
  if  opts.T_genkey == "Random-Different":
    T_RandomPermKey = True
    T_SamePermKeys=False
  elif  opts.T_genkey == "Random-Same":
    T_RandomPermKey = True
    T_SamePermKeys=True
  elif  opts.T_genkey == "Ident-Different":
    T_RandomPermKey = True
    T_SamePermKeys=False
  elif  opts.T_genkey == "Ident-Same":
    T_RandomPermKey = False
    T_SamePermKeys=True
  else:
    print("Error with -g: Argument \"", opts.T_genkey, "\" is not allowed.")
    exit()

# Set key length: -l
if opts.keylen:
  lop = opts.keylen
  if lop < 0 or lop > 256:
    print("Error with -l: Given key length must not be negative or bigger than 256.")
    exit()
  if lop!=256 and not(opts.first) and not(opts.T_genkey):
    print("Error: Setting the key length requires to specify option -f too (when doing enc or dec).")
    exit() 

# Raise exception: not -e, not -d and not -g
if not (opts.T_EncryptFileArg or opts.T_DecryptFileArg or opts.T_genkey):
  print("Error: Please enter a valid option: -e, -d, -g.")
  exit()


## Interpret options, set according global variables, process options
## (if more than one option is involved)

if T_verbose and T_firstAlphChar:
  print(first)



## Generate a key and exit (only when -g)
## - create the key (it consists of two permutations which have the same length lop)
## - each permutation is either a random permutation or an identical permutation
## - if T_SamePermKeys==True both newly created perms are the same.
if T_WritePermKeysToFile==True:
  permS=CreatePermKey(lop, T_RandomPermKey)
  if T_SamePermKeys==False:
    permT=CreatePermKey(lop, T_RandomPermKey)
  else:
    permT=permS
  f_permout=open(Name_f_permout, "wb")
  f_permout.write(bytearray(permS))
  f_permout.write(bytearray(permT))
  f_permout.close()
  if T_RandomPermKey==True and T_SamePermKeys==False:
    print( "Two different random perms, each of length", lop, "written as key to file", Name_f_permout, ".")
  elif T_RandomPermKey==False and T_SamePermKeys==False:
    print( "Two different identical perms, each of length", lop, "written as key to file", Name_f_permout, ".")
  elif T_RandomPermKey==True and T_SamePermKeys==True:
    print( "Two same random perms, each of length", lop, "written as key to file", Name_f_permout, ".")
  elif T_RandomPermKey==False and T_SamePermKeys==True:
    print( "Two same identical perms, each of length", lop, "written as key to file", Name_f_permout, ".")
  exit()



### Main Program
### - performs encryption or decryption (only when -e or -d).
### - reads the key and the input (plaintext or ciphertext) each from a file.
### ============

# Create physical storage for a list to be used for generating further storage (via slicing)
stor=[]
for i in range(lop):
  stor.append(i)

# Allocate memory for the two permutations
permS=stor[:]
permT=stor[:]

# Read key from file
if T_ReadPermKeysFromFile==True:
  f_permin =open(Name_f_permin, "rb")
  permS=f_permin.read(lop)
  permT=f_permin.read(lop)
  f_permin.close()
  permS= list(permS) 
  permT= list(permT) 
  if len(permS+permT)!= 2*lop:
    print("Info: Length of key file = " , len(permS+permT), ".")
    if lop != 256:
      print("Error: Parameter from option -l (", lop, ") is wrong. It should be half of key length.")
    else:
      print("Error: Length of key file does not match to the expected length of 2*256 bytes.")
    exit()
  if T_DecryptFileArg == True:
    permS=InversePerm(lop, permS)
    permT=InversePerm(lop, permT)

  if T_verbose:
    print("permS:     ", permS)
    print("permT:     ", permT)

# Allocate memory for one data blocks
DataOut=stor[:]

# Open input and output file
f_in  = open(Name_f_in, "rb")
f_out = open(Name_f_out, "wb")

# Length of the output file (changed during the main loop)
kkoF = 0


## Main loop: Process each block with length lop until EOF
while True:
  # Read from data intput file
  DataIn = bytearray(f_in.read(lop)) # bytearray (read/write) rather than bytes (read-only).

  lenDataIn = len(DataIn)
  kkoF += lenDataIn
  if T_verbose:
    print("Read Data=", DataIn, "len=", lenDataIn)

  # Block length == 0: This happens at the end when input length is a multiple of lop
  if lenDataIn == 0:
    break


  # Padding if: 0 < block length of read data < lop
  # Padded value must be part of the used alphabet (first, ..., first+lop-1)
  #   ( if not: problems with index in EncDecSubst(), DataOut[i]=perm[DataIn[i]] )
  # Padding: Default: with X=88 (only if X is part of the alphabet)
  #          Else   : with the first character of the alphabet.
  if lenDataIn < lop and lenDataIn > 0:
    ValueCharX = ord('X')
    if ValueCharX >= first and ValueCharX < lop+first:
      DataIn=DataIn + (lop-lenDataIn)*b"X"
    else:
	    # DataIn=DataIn + (lop-lenDataIn)*b'\x00'  # Less flexible alternative: Fill up with zeros
      OneString = (lop-lenDataIn)*chr(first)
      DataIn=DataIn + bytearray(OneString.encode())
    if T_verbose:
      print("Extended Data=", DataIn, ", len=", len(DataIn))
    kkoF += lop-lenDataIn


  # Now: Length of given block length == lop 
  # Within given input: Map alphabet down to range 0 to lop-1
  # (only necessary, if option -l is used; default uses complete range of a byte)
  if T_firstAlphChar:
    iTemp = 10
    for i in range(lop):
      iTemp=DataIn[i]-first  # Don't assign directly to DataIn[i] (as if input contains chars < first,
                             # then negative values are calculated, causing exception when assigned to bytes)
      # print("iTemp=", iTemp)   # for Debugging only
      if iTemp >= lop:
        print("Error: Inputdata out of range (bigger than lop=", lop, "): ", iTemp)
        exit()      
      if iTemp < 0:
        print("Error: Inputdata out of range (negative): first = ", first, ", iTemp = ", iTemp)
        if DataIn[i] == ord('\r') or DataIn[i] == ord('\n'):
          print("Error: Newline detected. This is not part of the alphabet. File may only contais characters which are in alphabet!")
        exit()
      DataIn[i]=iTemp


  # Perform encryption or decryption for this block.
  EncDecSubst(lop, permS, DataIn, DataOut)
  if T_verbose:
    print("SDaten:     ", DataOut)

  DataIn = DataOut[:]
  EncDecTranspos(lop, permT, DataIn, DataOut)
  if T_verbose:
    print("TDaten:     ", DataOut)


  # Created output: Map alphabet back to range first to first+lop-1
  # (only necessary, if option -l is used; default uses complete range of a byte)
  if T_firstAlphChar:
    for i in range(lop):
      DataOut[i]=DataOut[i]+first      

  f_out.write(bytearray(DataOut)) 


  # Leave while loop
  if lenDataIn < lop and lenDataIn > 0:
    break



# Clean up
f_in.close()
f_out.close()

# Notify user
if T_EncryptFileArg==True:
  print("Ciphertext of length", kkoF, " written to ", Name_f_out, ".")
if T_DecryptFileArg==True:
  print("Decrypted ciphertext of length", kkoF, " written to ", Name_f_out, ".")

